//
// apa.io.c
//
// APA I/O node
//
// this modified version made by
// Takuma Oami
// Fablab Vestmannaeyjar 2014
//
// set lfuse to 0x7E for 20 MHz xtal (for now just use 0x5E as default ... T)
//
// original code is wrote by
// Neil Gershenfeld
// CBA MIT 7/3/11
//
// (c) Massachusetts Institute of Technology 2011
// Permission granted for experimental and personal use;
// license for commercial sale available from MIT.
//

#include <apa.h>

// PWM for BLUE LED 
#define PWM_pin (1 << PD5)
#define PWM_port PORTD
#define PWM_direction  DDRD

// GREEN LED as connection checker 
#define LED_pin (1 << PD3)
#define LED_port PORTD
#define LED_direction  DDRD

uint8_t apa_hex_int2(unsigned char *ptr) {
  //
  // convert hex word to int
  //
  unsigned char hex_char(unsigned char arg) {
    if (('0' <= arg) && (arg <= '9'))
      return (arg - '0');
    else if (('a' <= arg) && (arg <= 'f'))
      return (10 + arg - 'a');
    else if (('A' <= arg) && (arg <= 'F'))
      return (10 + arg - 'A');
    else
      return 0;
  }
  return ((hex_char(ptr[0]) << 4) + hex_char(ptr[1]));
}


//
// process the packet
//
void apa_process_packet(struct apa_port_type *port) {
   uint8_t pwm;
   //unsigned char pwml, pwmh;
   static unsigned int cnt;
   //
   // execute command
   //
   switch (port->payload_out[0]) {
      case 'u':
         //
         // u: PWM up
         //
         pwm = OCR0B;
         pwm += -1;
         OCR0B = pwm;
         port->payload_out[0] = pwm;
         port->payload_out_length = 1;
         break;
      case 'd':
         //
         // d: PWM down
         //
         pwm = OCR0B;
         pwm += 1;
         OCR0B = pwm;
         port->payload_out[0] = pwm;
         port->payload_out_length = 1;
         break;
      case 'n':
         //
         // n: PWM on
         //
         OCR0B = 0;
	 pwm = 0;
         port->payload_out[0] = pwm;
         port->payload_out_length = 1;
         break;
      case 'f':
         //
         // f: PWM off
         //
         OCR0B = 255;
         pwm = 255;
         port->payload_out[0] = pwm;
         port->payload_out_length = 1;
         break;
      case 'r':
         //
         // r: PWM read
         //
         pwm = OCR0B;
         port->payload_out[0] = pwm;
         port->payload_out_length = 1;
         break;
      case 'w':
         //
         // w: PWM write
         //   wxx writes 8-bit hex value
         //
         pwm = apa_hex_int2(&port->payload_out[1]);
         OCR0B = pwm;
         port->payload_out[0] = pwm;
         port->payload_out_length = 1;
         break;
      case 's':
	 //
	 // sendback reverse path
	 //
	 cnt = 0;
	 for (cnt = 0; cnt < port->path_out_length; ++cnt){
	   if(port->path_out[cnt] == apa_pointer){
	     port->payload_out[cnt] = port->id;
	   }else{
	     port->payload_out[cnt] = port->path_out[cnt];
	   }
	 }
	 port->payload_out_length = port->path_out_length;
	 break;
      case ' ':
         //
         // do nothing
         //
         port->path_out_length = 0;
         port->payload_out_length = 0;
      default:
         //
         // unknown command
         //
         port->payload_out[0] = '?';
         port->payload_out_length = 1;
         break;
      }
   }

//
// main
//

int main(void) {
  static struct apa_port_type port_0, port_1, port_2, port_3, port_4, port_5;
   //
   // set clock divider to /1
   //
   CLKPR = (1 << CLKPCE);
   CLKPR = (0 << CLKPS3) | (0 << CLKPS2) | (0 << CLKPS1) | (0 << CLKPS0);
   //
   // initialize PWM counter and pin
   //
   TCCR0A = ((1 << COM0B1) | (1 << COM0B0) | (1 << WGM01)
	     | (1 << WGM00)); // clear OC1B on compare match, fast 8-bit PWM
   TCCR0B = ((0 << WGM02) | (0 << CS02) | (0 << CS01) 
	     | (1 << CS00));// /1 clock divider
   OCR0B = 255;
   set(PWM_port, PWM_pin);
   output(PWM_direction, PWM_pin);
   //
   // initialize ports and pins
   //
   port_0.pins_in = &PINB;
   port_0.port_out = &PORTD;
   port_0.direction_out = &DDRD;
   port_0.pin_in = (1 << PB0);
   port_0.pin_out = (1 << PD7);
   port_0.path_in_length = 0;
   port_0.path_out_length = 0;
   port_0.id = '0';
   port_0.next_port = &port_1;
   clear(*port_0.port_out, port_0.pin_out);
   output(*port_0.direction_out, port_0.pin_out);
   //
   port_1.pins_in = &PINC;
   port_1.port_out = &PORTC;
   port_1.direction_out = &DDRC;
   port_1.pin_in = (1 << PC5);
   port_1.pin_out = (1 << PC4);
   port_1.path_in_length = 0;
   port_1.path_out_length = 0;
   port_1.id = '1';
   port_1.next_port = &port_2;
   clear(*port_1.port_out, port_1.pin_out);
   output(*port_1.direction_out, port_1.pin_out);
   //
   port_2.pins_in = &PIND;
   port_2.port_out = &PORTD;
   port_2.direction_out = &DDRD;
   port_2.pin_in = (1 << PD0);
   port_2.pin_out = (1 << PD1);
   port_2.path_in_length = 0;
   port_2.path_out_length = 0;
   port_2.id = '2';
   port_2.next_port = &port_3;
   clear(*port_2.port_out, port_2.pin_out);
   output(*port_2.direction_out, port_2.pin_out);
   //
   port_3.pins_in = &PINB;
   port_3.port_out = &PORTB;
   port_3.direction_out = &DDRB;
   port_3.pin_in = (1 << PB2);
   port_3.pin_out = (1 << PB1);
   port_3.path_in_length = 0;
   port_3.path_out_length = 0;
   port_3.id = '3';
   port_3.next_port = &port_4;
   clear(*port_3.port_out, port_3.pin_out);
   output(*port_3.direction_out, port_3.pin_out);
   //
   port_4.pins_in = &PINC;
   port_4.port_out = &PORTC;
   port_4.direction_out = &DDRC;
   port_4.pin_in = (1 << PC0);
   port_4.pin_out = (1 << PC1);
   port_4.path_in_length = 0;
   port_4.path_out_length = 0;
   port_4.id = '4';
   port_4.next_port = &port_5;
   clear(*port_4.port_out, port_4.pin_out);
   output(*port_4.direction_out, port_4.pin_out);
   //
   port_5.pins_in = &PINC;
   port_5.port_out = &PORTC;
   port_5.direction_out = &DDRC;
   port_5.pin_in = (1 << PC2);
   port_5.pin_out = (1 << PC3);
   port_5.path_in_length = 0;
   port_5.path_out_length = 0;
   port_5.id = '5';
   port_5.next_port = &port_0;
   clear(*port_5.port_out, port_5.pin_out);
   output(*port_5.direction_out, port_5.pin_out);
   //
   // power on delay
   //
   power_on_delay();
   //
   //Turn on connection checker
   //
   clear(LED_port, LED_pin);
   output(LED_direction, LED_pin);

   //
   // main loop
   //
   while (1) {
      apa_port_scan(&port_0);
      apa_port_scan(&port_1);
      apa_port_scan(&port_2);
      apa_port_scan(&port_3);
      apa_port_scan(&port_4);
      apa_port_scan(&port_5);
      }
   }
